// @ myth - common - parse.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "common/parse.h"
#include "common/error.h"
#include "common/format.h"

// @@ class PARSE

// @@@ void PARSE::wrong(const char* where, const char* what)

void PARSE::wrong(const char* where, const char* what)
{
    bug.warning(format("%s line %d in %s", where, CurrentLine, FileName()),
		what);
}

// @@@ void PARSE::wrong(const char* where, STRING what)

void PARSE::wrong(const char* where, STRING what)
{
    wrong(where, what());
}

// @@@ char PARSE::get()

char PARSE::get()
{
    char tmp;

    if(NextChar == EOF)
    {
	BUFFER::get(&tmp, sizeof(char));
	if(tmp == '\n')
	    CurrentLine++;
    }
    else
    {
	tmp = NextChar;
	NextChar = EOF;
    }
    return tmp;
}

// @@@ void PARSE::put(char c)

void PARSE::put(char c)
{
    if(NextChar != EOF)
	wrong("PARSE::put", format("NextChar = %d, c = %c", NextChar, c));
    NextChar = c;
}

// @@@ bool PARSE::LookAhead(input_type type, char c = EOF)

bool PARSE::LookAhead(input_type type, char c = EOF)
{
    if(c == EOF)
    {
	c = get();
	put(c);
    }
    switch(type)
    {
    case INPUT_WORD:
    case INPUT_LINE:
	return TRUE;
    case INPUT_RANGE:
	return isdigit(c) || c == '.';
    case INPUT_LONG:
	return isdigit(c) || c == '-';
    case INPUT_SPACE:
	return c == ' ' || c == '\t';
    case INPUT_WHITE:
	return strchr(" \t\n", c) != NULL;
    case INPUT_NEWLINE:
	return c == '\n';
    case INPUT_COMMENT:
	return c == '\n' || c == '#';
    case INPUT_KEY:
    case INPUT_STRING:
	return c == '"';
    default:
	wrong("PARSE::LookAhead", 
		  format("Unknown type %d at '%c'", type, c));
	return FALSE;
    }
}

// @@@ bool PARSE::LookAhead(const char* s)

bool PARSE::LookAhead(const char* s)
{
    char c = get();
    put(c);
    return strchr(s, c) != NULL;
}    

// @@@ void PARSE::skip(input_type type)

void PARSE::skip(input_type type)
{
    switch(type)
    {
    case INPUT_WORD:
	int c;
	do
	    c = get();
	while(strchr(" \t\n", c) == NULL && good());
	put(c);
	return;
    case INPUT_LINE:
	do
	    c = get();
	while(c != '\n' && good());
	return;
    case INPUT_RANGE:
	skip(INPUT_LONG);
	if(!LookAhead(INPUT_WHITE))
	    skip("..");
	skip(INPUT_LONG);
	return;
    case INPUT_LONG:
	c = get();
	if(c == '-')
	    c = get();
	while(isdigit(c) && good())
	    c = get();
	put(c);
	return;
    case INPUT_SPACE:
	do
	    c = get();
	while((c == ' ' || c == '\t') && good());
	put(c);
	return;
    case INPUT_WHITE:
	do
	    c = get();
	while(strchr(" \t\n", c) != NULL && good());
	put(c);
	return;
    case INPUT_NEWLINE:
	do
	    skip(INPUT_LINE);
	while(LookAhead(INPUT_NEWLINE) && good());
	return;
    case INPUT_COMMENT:
	do
	    skip(INPUT_LINE);
	while(LookAhead(INPUT_COMMENT) && good());
	return;
    case INPUT_KEY:
    case INPUT_STRING:
	if(c != '"')
	    return;

	bool escape = FALSE;
	while(good())
	{
	    if(escape)
	    {
		(void) get();
		escape = FALSE;
	    }
	    else
	    {
		switch(get())
		{
		case '\\':
		    escape = TRUE;
		case '"':
		    return;
		}
	    }
	}
	return;
    default:
	wrong("PARSE::skip", format("Unknown type %d", type));
    }
}

// @@@ void PARSE::skip(const char* s)

void PARSE::skip(const char* s)
{
    while(good() && *s)
    {
	if(*s == get())
	    s++;
	else
	{
	    wrong("PARSE::skip()", format("'%c' unmatched", *s));
	    break;
	}
    }
}

// @@@ char* PARSE::GetWord()

char* PARSE::GetWord()
{
    int pos = 0;
    char c;
    
    while(good())
    {
	c = get();
	if(LookAhead(INPUT_WHITE, c))
	    break;
	if(pos < MessageSize - 1)
	    buffer[pos++] = c;
    }
    buffer[pos] = 0;
    put(c);

    return buffer;
}

// @@@ char* PARSE::GetLine()

char* PARSE::GetLine()

{
    int pos = 0;
    char c;
    
    while(good())
    {
	c = get();
	if(LookAhead(INPUT_NEWLINE, c))
	    break;
	if(pos < MessageSize - 1)
	    buffer[pos++] = c;
    }
    buffer[pos] = 0;

    return buffer;
}

// @@@ range PARSE::GetRange()

range PARSE::GetRange()
{
    if(LookAhead(INPUT_LONG))
    {
	long first = GetLong();
	if(LookAhead(INPUT_WHITE))
	    return range(first, first);
	else
	{
	    skip("..");
	    if(LookAhead(INPUT_LONG))
		return range(first, GetLong());
	    else
		return range(first, INFINITE);
	}
    }
    else
    {
	skip("..");
	if(LookAhead(INPUT_LONG))
	    return range(INFINITE, GetLong());
    }
    return range(INFINITE, INFINITE);
}

// @@@ long PARSE::GetLong()

long PARSE::GetLong()
{
    long value = 0L;
    bool negative = FALSE;
    char c = get();

    while(good())
    {
	if(c == '-')
	    negative = TRUE;
	else if(isdigit(c))
	    value = value * 10 + c - '0';
	else
	    break;

	c = get();
    }
    put(c);

    return negative ? -value : value;
}

// @@@ char PARSE::GetKey()

char PARSE::GetKey()
{
    char c = get();

    switch(c)
    {
    case '^':
	return control(get());
    case '\\':
	return get();
    default:
	return c;
    }
}

// @@@ char* PARSE::GetString()

char* PARSE::GetString()
{
    skip("\"");

    int pos = 0;
    char c = GetKey();
    while(c != '"')
    {
	if(pos < MessageSize - 2)
	    buffer[pos++] = c;
	c = GetKey();
    }
    buffer[pos] = 0;

    return buffer;
}

// @@@ PARSE::PARSE(STRING name)

PARSE::PARSE(STRING name) : (BUFFER_READ, name)
{
    FileName = name;
    CurrentLine = 1;
    NextChar = EOF;
    
    if(BUFFER::GoodInput() == ERROR)
	bug.warning("PARSE::PARSE", format("Can't read %s", name()));
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
