// @ myth - common - wait.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk
//
// This file contain the class WAIT.
//
// WAIT provide a way to check a number of buffers (of the type
// BUFFER), to see if there is any data ready.
//
// To use it, first define an object of the type WAIT
//    WAIT wait;
//
// Then tell WAIT what buffers you want to test
//    wait.add(SomeBuffer);
//    wait.add(AnotherBuffer);
//    etc...
//
// Each buffer contain a flag, IsReady, which will be set by
// wait.check(), if the buffer contain data ready to read.
//    bool TimeOut = wait.check();
//    if(SomeBuffer->IsReady)
//        ReadDataFromSomeBuffer...
// The IsReady-flag will only be updated by wait.check(), thus
// wait.check() should be called each time a byte is read from the
// buffer, if you want to avoid blocking i/o totally.
//
// When you don't want to check a buffer any more, call
//    wait.remove(SomeBuffer);
//
// WARNING:
//    This class won't work if the system supports more than 32
//    filedescriptors.
//
// It shouldn't be to difficult to port to SysV.

// @@ include

#include "common/wait.h"
#include "common/error.h"
#include "common/format.h"

// @@ * class WAIT

// @@@ * void WAIT::set(int fd)

void WAIT::set(int fd)
{
    int index = 0;

    while(fd > 31)
    {
	fd -= 32;
	index++;
    }
    files[index] |= 1 << fd;
}

// @@@ * void WAIT::clear(int fd)

void WAIT::clear(int fd)
{
    int index = 0;

    while(fd > 31)
    {
	fd -= 32;
	index++;
    }
    files[index] &= ~(1 << fd);
}

// @@@ * bool WAIT::test(int* mask, int fd)

bool WAIT::test(int* mask, int fd)
{
    int index = 0;

    while(fd > 31)
    {
	fd -= 32;
	index++;
    }
    return mask[index] & (1 << fd);
}

// @@@ * void WAIT::add(BUFFER* buf)
//
// Add a buffer to the wait queue.

void WAIT::add(BUFFER* buf)
{
    list = new buffer_link(buf, list);
    set(buf->fileno());
}

// @@@ * void WAIT::remove(BUFFER* buf)
//
// Remove a buffer from the wait queue.

void WAIT::remove(BUFFER* buf)
{
    buffer_link* previous = NULL;
    
    for(buffer_link* current = list; current != NULL; current = current->next)
    {
	if(current->buffer == buf)
	{
	    if(previous == NULL)
		list = current->next;
	    else
		previous->next = current->next;

	    clear(buf->fileno());
	    delete current;
	    return;
	}
	previous = current;
    }
    bug.error("WAIT::remove", "buf not found");
}	

// @@@ * void WAIT::check(int timeout = ERROR)
//
// Update IsReady for some buffers in the wait queue.
//
// The optional argument may be:
//    ERROR:	Wait until some buffer has data ready.
//    		May not update other buffers.
//    0:	Update all buffers.
//    arg > 0:	Update all buffers, then wait until some buffer has data
//    		ready or wait for arg seconds, whatever come first.

void WAIT::check(int timeout = ERROR)
{
    // Examine the content of the buffers
    bool AllReady = TRUE;	// TRUE, iff all buffers contain data
    bool SomeReady = FALSE;	// TRUE, iff any buffer contain data

    for(buffer_link* link = list; link != NULL; link = link->next)
    {
	if(link->buffer->empty())
	    link->buffer->IsReady = AllReady = FALSE;
	else
	    link->buffer->IsReady = SomeReady = TRUE;
    }
    if(AllReady || (timeout == ERROR && SomeReady))
	return;			// No reason to call select.
    
    // Create timeout value
    timeval time;

    if(timeout != ERROR)
    {
	time.tv_sec = SomeReady ? 0 : timeout;
	time.tv_usec = 0L;
    }

    for(int i = 0; i < length; i++)
	mask[i] = files[i];
    
    // Examine file descriptors.
    switch(select(table, mask, NULL, NULL, (timeout == ERROR) ? NULL : &time))
    {
    case ERROR:
	bug.error("WAIT::check",
		  format("select(%d, %d, ...) failed", table, mask[0]));
	break;
    case 0:			// Time Out
	break;
    default:
	for(link = list; link != NULL; link = link->next)
	    if(test(mask, link->buffer->fileno()))
		link->buffer->IsReady = TRUE;
    }
    return;
}

// @@@ WAIT::WAIT()

WAIT::WAIT()
{
    list = NULL;
    table = getdtablesize();
    length = (table - 1) / 32 + 1;
    files = new int[length];
    for(int i = 0; i < length; i++)
	files[i] = 0;
    mask = new int[length];
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
